import funcy as fn
import peewee as pw


__all__ = (
    "model_serializer",
)


def model_serializer(
    model_class, fields=None, follow_m2m=True, exclude=None, extra=None, **serializers
):
    fields = fields or model_class._meta.fields.keys()
    if extra:
        fields = fields | extra

    if follow_m2m:
        fields = fields | model_class._meta.manytomany.keys()

    if exclude:
        fields = fields - set(exclude)

    def get_field_value(obj, field_name):
        if field_name in model_class._meta.fields:
            field = model_class._meta.fields[field_name]
        elif field_name in model_class._meta.manytomany:
            field = model_class._meta.manytomany[field_name]
        else:
            field = None

        if field and getattr(obj, field_name) is not None:
            if field.__class__ in (pw.DateField, pw.DateTimeField):
                return getattr(obj, field_name).isoformat()
            elif field.__class__ == pw.ForeignKeyField:
                return model_serializer(field.rel_model, follow_m2m=False)(
                    getattr(obj, field_name)
                )
            elif field.__class__ == pw.ManyToManyField:
                return model_serializer(field.rel_model, follow_m2m=False)(
                    getattr(obj, field_name)
                )
            elif field.__class__ == pw.DecimalField:
                return str(getattr(obj, field_name))

        return getattr(obj, field_name)

    defaults = dict(
        fn.lmap(
            lambda field_name: (
                field_name,
                lambda obj: get_field_value(obj, field_name),
            ),
            fields,
        )
    )
    serializers = fn.merge(defaults, serializers)

    def object_serializer(obj, project=None, omit=None):
        fields = serializers.keys()
        if project:
            fields = fields & project
        if omit:
            fields = fields - omit
        return dict([(f, serializers[f](obj)) for f in fields])

    def inner(qs, project=None, omit=None, **_fields):
        if isinstance(qs, pw.Model):
            return object_serializer(qs, project=project, omit=omit)
        return [object_serializer(obj, project=project, omit=omit) for obj in qs]

    return inner



import funcy as fn
import requests

__all__ = ("ApiWrapper", "api_headers",)


class HeaderRegistry(list):
    def __call__(self):
        def func_decorator(func):
            self.append(func)
            return func
        return func_decorator

api_headers = HeaderRegistry()

class ApiWrapper(object):
    host = None
    methods = ["get", "post", "put", "delete", "head", "patch", "options"]

    def __init__(self):
        for method in self.methods:
            setattr(self, method, fn.partial(self.call, method))

    def get_headers(self):
        headers = {}
        for api_header_fn in api_headers:
            api_header_fn(headers)
        return headers

    def call(self, method, uri, *args, **kwargs):
        if not self.host:
            raise RuntimeError("Host not defined")

        kwargs["headers"] = fn.merge(kwargs.get("headers", {}), self.get_headers())
        return getattr(requests, method)(f"{self.host}{uri}", *args, **kwargs)



import math
import logging
import datetime
import decimal
import types
import funcy as fn

import json
import peewee as pw

from webob import Response, exc
from playhouse.shortcuts import model_to_dict, dict_to_model

from .http import *
from .db import *

from functools import reduce
import operator

is_str = fn.isa(str)

__all__ = (
    "get_object_or_404",
    "get_object_or_400",
    "get_object_or_422",
    "paginated",
    "filtered",
    "ordered",
    "skimmed",
    "object_list",
    "object_detail",
)



def get_object_or_404(model_class, **kwargs):
    obj = model_class.get_or_none(**kwargs)
    if not obj:
        not_found(error="{} Not found".format(model_class.__name__))
    return obj


def get_object_or_400(model_class, **kwargs):
    obj = model_class.get_or_none(**kwargs)
    if not obj:
        bad_request(error="{} Not found".format(model_class.__name__))
    return obj


def get_object_or_422(model_class, **kwargs):
    obj = model_class.get_or_none(**kwargs)
    if not obj:
        unprocessable_entity(error="{} Not found".format(model_class.__name__))
    return obj


def paginated(req, qs, paginate_by=15, max_paginate_by=100):
    page = int(req.params.get("page", 1))
    paginate_by = min(int(req.params.get("paginate_by", paginate_by)), max_paginate_by)
    record_count = qs.count()
    page_count = int(math.ceil(record_count / paginate_by))
    return (
        qs.paginate(page, paginate_by),
        {
            "is_paginated": record_count > paginate_by,
            "page": page,
            "page_count": page_count,
            "record_count": record_count,
        },
    )


VALUE_CONVERSION = {"true": True, "false": False, "none": None}



@fn.memoize
def get_model_fields(model_class):
    return [f for f in model_class._meta.fields.keys()]


def filtered(req, qs, allowed_fields=None, omit_fields=None, expr_fns=None, **fns):

    params = req.params.mixed()

    filters = dict(
        fn.lmap(
            lambda x: (fn.cut_prefix(x[0], "filter__"), x[1]),
            fn.filter(lambda x: x[0].startswith("filter__"), fn.iteritems(params)),
        )
    )

    if filters:
        model_class = qs.model
        model_fields = get_model_fields(model_class)

        if allowed_fields:
            model_fields = [f for f in model_fields if f in allowed_fields]

        if omit_fields:
            model_fields = [f for f in model_fields if f not in omit_fields]

        expr_fns = expr_fns or {}

        _expr_list = []

        for fkey, value in filters.items():
            tokens = fkey.split("__")

            if len(tokens) == 1:
                field = tokens[0]
                op = "eq"
            elif len(tokens) == 2:
                field = tokens[0]
                op = tokens[1]

            # coerce values
            if op in ["in", "not_in"]:
                value = value if fn.is_list(value) else [value]
            elif is_str(value):
                value = VALUE_CONVERSION.get(value, value)

            if field in fns:
                qs = fns[field](req, qs, value)
                continue

            if field in expr_fns:
                _expr_list.append(expr_fns[field](value))
                continue

            if field not in model_fields:
                bad_request(error=f"field {field} not filterable")

            if op == "eq":
                _expr_list.append(getattr(model_class, field) == value)
            elif op == "in":
                _expr_list.append(getattr(model_class, field).in_(value))
            elif op == "not_in":
                _expr_list.append(getattr(model_class, field).not_in(value))
            elif op == "is_null":
                # FIXME assert boolean
                _expr_list.append(getattr(model_class, field).is_null(value))
            elif op == "contains":
                _expr_list.append(getattr(model_class, field).contains(value))
            else:
                bad_request(error=f"field: {field} unknown operation: {op}")

        if _expr_list:
            qs = qs.where(reduce(operator.and_, _expr_list))


    return qs


def ordered(req, qs, **order_fns):
    for order in req.params.getall("order_by"):
        direction = "desc" if order[0] == "-" else "asc"
        field_name = order[1:] if order[0] in ("+", "-") else order

        if field_name in order_fns:
            qs = order_fns[field_name](req, qs, direction)
        else:
            field = getattr(qs.model, field_name)
            if direction == "desc":
                field = field.desc()
            qs = qs.order_by(field)
    return qs


def skimmed(req, qs, serializer):
    return serializer(
        qs,
        project=req.params.getall("field") or None,
        omit=req.params.getall("exclude") or None,
    )


def object_list(
    req,
    qs,
    serializer=None,
    key_name=None,
    filter_kwargs=None,
    order_fns=None,
    paginate_by=15,
    max_paginate_by=100,

):
    serializer = serializer or qs.model.serializer()
    qs = filtered(req, qs, **(filter_kwargs or {}))
    qs = ordered(req, qs, **(order_fns or {}))
    (qs, pagination) = paginated(
        req, qs, paginate_by=paginate_by, max_paginate_by=max_paginate_by
    )
    key_name = key_name or f"{qs.model.__name__.lower()}_list"
    return {key_name: skimmed(req, qs, serializer), "pagination": pagination}


def object_detail(req, obj, serializer=None, key_name=None):
    serializer = serializer or obj.__class__.serializer()
    key_name = key_name or obj.__class__.__name__.lower()
    return {key_name: skimmed(req, obj, serializer)}



import funcy as fn
import msgpack
from walrus import *
from .appconfig import appconfig
from .settings import settings

__all__ = ("cachedb", "cache", "evict")

is_str = fn.isa(str)


@appconfig.settings()
def cache_settings(**opts):
    return {
        "cache_host": appconfig.env.str("REDIS_HOST", "localhost"),
        "cache_port": appconfig.env.int("REDIS_PORT", 6379),
        "cache_db": appconfig.env.int("CACHE_DB", 0),
        "cache_lock_duration": appconfig.env.int("CACHE_LOCK_DURATION", 500),  # in milliseconds
    }


@fn.LazyObject
def cachedb():
    return Database(
        host=settings.cache_host,
        port=settings.cache_port,
        db=settings.cache_db
    )


@fn.decorator
def cache(call, *, key=None, evict_keys=None):
    req = call._args[0]
    key = (
        key
        if is_str(key)
        else key(call)
        if callable(key)
        else ":".join(
            [call._func.__name__]
            + fn.lmap(lambda x: f"{x[0]}-{x[1]}", req.params.items())
            + fn.lmap(lambda x: f"{x[0]}-{x[1]}", call._kwargs.items())
        )
    )
    key = f"catalog:cache:{key}"
    resp = cachedb.get(key)
    if resp:
        return msgpack.unpackb(resp, raw=False)

    lock = cachedb.lock(key, ttl=settings.cache_lock_duration)
    with lock:
        resp = call()
        cachedb[key] = msgpack.packb(resp, use_bin_type=True)
        evict_keys = (
            evict_keys
            if fn.is_list(evict_keys)
            else [evict_keys]
            if is_str(evict_keys)
            else evict_keys(call)
            if callable(evict_keys)
            else []
        )
        evict_keys = [ek.format(**call._kwargs) for ek in evict_keys]
        for evict_key in evict_keys:
            cache_set = cachedb.Set(f"catalog:eviction:{evict_key}")
            cache_set.add(key)

    return resp


@fn.decorator
def evict(call, *evict_keys):
    resp = call()
    for evict_key in evict_keys:
        evict_key = evict_key(call) if callable(evict_key) else evict_key
        evict_key = evict_key.format(**call._kwargs)
        cache_set = cachedb.Set(f"catalog:eviction:{evict_key}")
        for key in cache_set.members():
            cachedb.delete(key)
        cache_set.clear()
    return resp



import logging
from pibe_ext.validator import *
from pibe_ext.http import http, no_content

logger = logging.getLogger(__name__)

__all__ = ("rpc",
        "JSONRPCParseError", "JSONRPCInvalidRequest", "JSONRPCMethodNotFound",
        "JSONRPCInvalidParams", "JSONRPCInternalError", "JSONRPCServerError",
)

# https://www.jsonrpc.org/specification

class JSONRPCException(Exception):

    def __init__(self, message):
        self.message = message

class JSONRPCParseError(JSONRPCException):
    code = -32700

class JSONRPCInvalidRequest(JSONRPCException):
    code = -32600

class JSONRPCMethodNotFound(JSONRPCException):
    code = -32601

class JSONRPCInvalidParams(JSONRPCException):
    code = -32601

class JSONRPCInternalError(JSONRPCException):
    code = -32603

class JSONRPCServerError(JSONRPCException):
    code = -32000


class RPCRegistry(dict):
    def __call__(self, method_name=None):
        def func_decorator(func):
            self[method_name or func.__name__] = func
            return func

        return func_decorator

    def dispatch(self, wh_name, *args, **kwargs):
        for callback_fn in self.get(wh_name, []):
            callback_fn(*args, **kwargs)


rpc = RPCRegistry()


@http.post("/rpc")
def process_rpc(req):
    try:
        payload = req.json
    except:
        return {
            "jsonrpc": "2.0",
            "error": {"code": -32700, "message": "Parse error"}
        }

    rpc_method = payload.get("method")
    rpc_params = payload.get("params")
    rpc_id =  payload.get("id")

    resp = {
        "jsonrpc": "2.0",
        "id": rpc_id
    }

    func = rpc.get(rpc_method)
    if not func:
        resp["error"] = {"code": -32601, "message": "Method not found"}
        return resp

    args = rpc_params if type(rpc_params) == list else []
    kwargs = rpc_params if type(rpc_params) == dict else {}

    try:
        resp["result"] = rpc[rpc_method](*args, **kwargs)
    except JSONRPCException as exc:
        resp["error"] = {"code": exc.code, "message": exc.message}
    except:
        logger.exception(f"RPC Server Error while executing rpc method: {rpc_method}", exc_info=True)
        resp["error"] = {"code": -32000, "message": "Server Error"}

    return resp




import peewee as pw
import playhouse.postgres_ext as pw_pext

import funcy as fn
from webob import exc
from json.decoder import JSONDecodeError
import cerberus
from pibe import DotDict
import arrow

from .http import _raise_exc, not_acceptable

__all__ = (
    "model_schema",
    "model_validate"
)


FIELD_MAP = {
    'smallint': 'number',
    'bigint': 'number',
    'bool': 'boolean',
    'date': 'date',
    'datetime': 'datetime',
    'decimal': 'number',
    'double': 'number',
    'float': 'float',
    'int': 'integer',
    'time': 'string',
}

COERCE_MAP = {
    'float': float,
    'int': int,
    "date": lambda s: arrow.get(s).date(),
    "datetime": lambda s: arrow.get(s).datetime,
}

def dict_merge(dct, merge_dct):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k, v in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], dict)):  #noqa
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]

def model_schema(model_class, fields=None, exclude=None, schema=None):
    fields = fields or model_class._meta.fields.keys()
    exclude = exclude or []
    schema = schema or {}
    _schema = {}

    for name, field in model_class._meta.fields.items():
        if getattr(field, 'primary_key', False):
            continue
        if name in exclude:
            continue
        if fields and name not in fields:
            continue
        field_type = field.field_type.lower()
        required = not bool(getattr(field, 'null', True))
        if type(field) == pw_pext.ArrayField:
            ctype = "list"
        elif type(field) == pw_pext.HStoreField:
            ctype = "dict"
        else:
            ctype = FIELD_MAP.get(field_type, "string")

        field_schema = {
            "type": ctype,
            "required": True,
        }
        # field_schema["empty"] = not required
        choices = getattr(field, 'choices', ())
        if choices:
            field_schema["allowed"] = [c[0] for c in choices]
        default = getattr(field, 'default', None)
        max_length = getattr(field, 'max_length', None)
        if max_length:
            field_schema["maxlength"] = max_length

        unique = getattr(field, 'unique', False)
        coerce = COERCE_MAP.get(field_type)
        if coerce:
            field_schema["coerce"] = coerce
        _schema[name] = field_schema
    dict_merge(_schema, schema)
    return _schema


def model_validate(model_class,
    exception_class=exc.HTTPBadRequest,
    fields=None,
    exclude=None,
    schema=None,
    **kwargs):

    _schema = model_schema(model_class, fields=fields, exclude=exclude, schema=schema)

    def _deco(func):
        @fn.wraps(func)
        def wrapper(req, *args, **kwargs):
            try:
                data = dict(req.json)
            except JSONDecodeError:
                not_acceptable(error="Invalid JSON Request")

            v = cerberus.Validator(_schema, **kwargs)
            if not v.validate(data):
                _raise_exc(exception_class, errors=v.errors)

            req.data = DotDict(v.document)
            return func(req, *args, **kwargs)

        return wrapper
    return _deco



import gevent
import logging
from funcy import wraps

__all__ = ('greenlet', 'infinite_loop', 'main_loop')

logger = logging.getLogger(__name__)

def greenlet(**kwargs):
    later = kwargs.get('later', None)
    repeat = kwargs.get('repeat', None)
    immediate = kwargs.get('immediate', False)

    def inner_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if repeat:
                def _repeat(func, *args, **kwargs):
                    logger.debug('Started repeating function {}'.format(func))
                    try:
                        if immediate:
                            gevent.spawn(func, *args, **kwargs)

                        while 1:
                            gevent.sleep(repeat)
                            gevent.spawn(func, *args, **kwargs)
                    except gevent.GreenletExit:
                        logger.debug('Stopped repeating function {}'.format(func))
                        pass
                gl_args = (_repeat, func, *args)
            else:
                gl_args = (func, *args)

            if later:
                gl = gevent.spawn_later(later, *gl_args, **kwargs)
            else:
                gl = gevent.spawn(*gl_args, **kwargs)
            return gl
        return wrapper
    return inner_wrapper


@greenlet()
def infinite_loop():
    while 1:
        gevent.sleep(0)

def main_loop():
    try:
        gevent.joinall([infinite_loop()])
    except (KeyboardInterrupt, SystemExit):
        pass



import funcy as fn
from .settings import settings
from walrus import *

__all__ = ("rdb",)


@fn.LazyObject
def rdb():
    return Database(
        host=settings.redis_host,
        port=settings.redis_port,
        db=settings.redis_db
    )



import os
import json
import time
import funcy as fn
import threading
import queue
import logging
from redis import Redis

from pibe_ext.appconfig import *
from pibe_ext.settings import *
from pibe_ext.http import *
from pibe_ext.validator import *


__all__ = ("task",)

logger = logging.getLogger(__name__)

q = queue.Queue()

@appconfig.settings()
def task_settings(**opts):
    return {
        "tasks_sync_exec": appconfig.env.bool("TASKS_SYNC_EXEC", False),
        "tasks_redis_host": appconfig.env.str("REDIS_HOST", "localhost"),
        "tasks_redis_port": appconfig.env.int("REDIS_PORT", 6379),
        "tasks_redis_db": appconfig.env.bool("TASKS_REDIS_DB", 0),
        "tasks_redis_queue_name": appconfig.env("TASKS_REDIS_QUEUE_NAME", "tasks"),
    }


@fn.LazyObject
def tasksdb():
    return Redis(
        host=settings.tasks_redis_host,
        port=settings.tasks_redis_port,
        db=settings.tasks_redis_db
    )


class TaskWrapper(object):
    def __init__(self, func, task_name):
        self.func = func
        self.task_name = task_name

    def __call__(self, *args, **kwargs):
        if settings.tasks_sync_exec:
            self.func(*args, **kwargs)
        else:
            enqueue({
                "task": self.task_name,
                "args": args,
                "kwargs": kwargs,
            })


class TaskRegistry(dict):
    def __call__(self, name=None):
        def func_decorator(func):
            task_name = name or func.__name__
            self[task_name] = func
            return TaskWrapper(func, task_name)
        return func_decorator


task = TaskRegistry()


class Worker(threading.Thread):
    def __init__(self, q, n, *args, **kwargs):
        self.q = q
        self.n = n
        super().__init__(*args, **kwargs)

    def run(self):
        logger.info(f"[{threading.current_thread().name}] Starting Worker {self.n}")
        while True:
            try:
                payload = self.q.get(timeout=1)
            except queue.Empty:
                continue

            task_name = payload["task"]
            args = payload["args"]
            kwargs = payload["kwargs"]
            task_fn = task[task_name]
            try:
                t0 = time.time()
                task_fn(*args, **kwargs)
                elapsed = time.time() - t0
                logger.info(f"[{threading.current_thread().name}] [Worker {self.n}] [task={payload['task']} args={args} args={kwargs}] executed in {elapsed} seconds")
            except Exception as e:
                logger.exception(f"[{threading.current_thread().name}] [Worker {self.n}] [task={payload['task']} args={args} args={kwargs}] Error executing task.", exc_info=True)

            self.q.task_done()


class WorkerOrchestrator(threading.Thread):
    def __init__(self, q, *args, **kwargs):
        self.q = q
        super().__init__(*args, **kwargs)

    def run(self):
        logger.info(f"[{threading.current_thread().name}] Starting Worker Orchestrator [PID={os.getpid()}]")
        while True:
            msg = tasksdb.rpop(settings.tasks_redis_queue_name)
            if msg is None:
                time.sleep(0.1)
                continue

            payload = json.loads(msg)
            q.put_nowait(payload)


def runworkers(total_threads=3):
    threads = [WorkerOrchestrator(q)] + [Worker(q, n+1) for n in range(total_threads)]
    [t.start() for t in threads]
    [t.join() for t in threads]


def enqueue(payload):
    tasksdb.rpush(settings.tasks_redis_queue_name, json.dumps(payload))



import logging
from .settings import settings
from .appconfig import appconfig
from .http import http
from .session import g

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.wsgi import SentryWsgiMiddleware


@appconfig.settings()
def sentry_settings(**opts):
    return {
        "use_sentry": appconfig.env.bool("USE_SENTRY", False),
        "sentry_dsn": appconfig.env("SENTRY_DSN", None),
    }



@appconfig.initialize()
def initialize_sentry(**opts):
    if settings.use_sentry:
        sentry_sdk.init(
            dsn=settings.sentry_dsn,
            integrations=[
                LoggingIntegration(
                    level=settings.log_level, event_level=settings.log_level
                )
            ],
        )
        if settings.get("sentry_tags"):
            for k, v in settings.sentry_tags.items():
                sentry_sdk.set_tag(k, v)


@appconfig.wsgi_middleware()
def sentry_wsgi_middleware(application, **opts):
    if settings.use_sentry:
        return SentryWsgiMiddleware(application)
    return application


@http.before_request()
def sentry_middleware(req):
    if settings.use_sentry:
        sentry_sdk.set_tag("correlation_id", req.correlation_id)



import os
import logging
from environs import Env
import funcy as fn
from pibe_ext.settings import settings


__all__ = ("appconfig",)


class CallbackRegistry(list):
    def __call__(self):
        def func_decorator(func):
            self.append(func)
            return func
        return func_decorator


class AppConfig(object):

    def __init__(self):
        self.env = Env()
        self.env.read_env(os.environ.get("CONFIG_FILE", ".env"), recurse=False)
        self.settings = CallbackRegistry()
        self.initialize = CallbackRegistry()
        self.wsgi_middleware = CallbackRegistry()

    def _get_funcs(self, registry, **opts):
        funcs = [f for f in registry]

        only_fns = opts.get("only", [])
        if only_fns:
            funcs = [f for f in funcs if f.__name__ in only_fns]

        exclude_fns = opts.get("exclude", [])
        funcs = [f for f in funcs if f.__name__ not in exclude_fns]
        return funcs

    def init_settings(self, **opts):

        funcs = self._get_funcs(self.settings, **opts)

        if opts.get("dump_funcs", True):
            print("Collecting settings from {}".format(", ".join([f.__name__ for f in funcs])))

        settings.update(fn.merge(*[(f(**opts) or {}) for f in funcs]) or {})

    def init(self, **opts):
        if opts.get("initialize_settings", True):
            self.init_settings(**opts)

        funcs = self._get_funcs(self.initialize, **opts)

        if opts.get("dump_funcs", True):
            print("Initializing {}".format(", ".join([f.__name__ for f in funcs])))

        for func in funcs:
            func(**opts)

    def start_app(self, app, **opts):
        if opts.get("initialize", True) == True:
            self.init(**opts)

        if opts.get("install_middleware", True) == True:
            for func in self.wsgi_middleware:
                app = func(app, **opts)

        return app


appconfig = AppConfig()
